use strict;
use warnings FATAL => 'all';
use Excel::Writer::XLSX;
use Array::Utils qw(:all);
use List::MoreUtils qw[any uniq none distinct first_index];
use List::Util qw[shuffle max min sum];
use Scalar::Util qw(looks_like_number);
use Getopt::Long qw(GetOptions);
use Math::Gradient qw(array_gradient);
use Statistics::Descriptive;
use Sys::Hostname;

no if $] >= 5.017011, warnings => 'experimental::smartmatch';

$| = 1;

use constant
	{
	TRUE => "1",
	FALSE => "0",
	SAMPLE_MATRIX => 10000
	};

my $VERBOSE = FALSE;

main();

#############################################################################################################

sub main
	{
	############### parse command line arguments ###############

	my %user_arguments = parse_arguments();

	my %data = init_data(\%user_arguments);

	############### start Excel spreadsheet ###############

	my $workbook = Excel::Writer::XLSX -> new($data{'SPREADSHEET_FILE'});

	my $worksheet = $workbook->add_worksheet();
	
	############### defines colors and formats ###############

	my %color = define_colors($workbook, \%user_arguments, \%data);

	my %format = define_formats($workbook, \%user_arguments, \%data, %color);

	############### reads and processes matrix file ###############

	my @matrix = load_and_center_matrix(\%user_arguments, \%data);

	my %data_colors = read_colors_file(\%user_arguments, \%format, \%data);

	############### reads metadata (header) file and writes the header of the Excel spreadsheet above the matrix ###############

	%data = process_sampleinfo(\%user_arguments, \%data, $worksheet, \@matrix, \%data_colors, %format);

	############### reads matrix file and writes the matrix onto the Excel spreadsheet ###############

	%data = process_matrix_data(\%user_arguments, \%data, $worksheet, \@matrix, %format);

	############### reads gene info file and writes the info onto the Excel spreadsheet to the right side of the matrix ###############

	%data = process_geneinfo_data(\%user_arguments, \%data, \%data_colors, $worksheet, %format);

	############### final touches to the format of the Excel spreadsheet ###############

	format_output_file($worksheet, \%data);

	$workbook->close();

	############### final output messages ###############

	print "[OUTPUT] Results saved in '$user_arguments{'PREFIX'}.xlsx'\n" if ( ($user_arguments{'VERBOSE_FLAG'} eq "TRUE") and (-e "$user_arguments{'PREFIX'}.xlsx") );
	print "[ERROR] File '$user_arguments{'PREFIX'}.xlsx' cannot be saved.\n" if (!-e "$user_arguments{'PREFIX'}.xlsx");

	}

sub init_data
	{
	my $user_arguments_ref = $_[0];

	my %data = (
	'SPREADSHEET_FILE' => "$$user_arguments_ref{'PREFIX'}.xlsx",
	'SAMPLE_INFO_ROWS' => -1,
	'GENE_INFO_COLUMNS' => 0,
	'COLOR_LIST' => "WHITE MAGENTA LIME RED GREEN YELLOW BLUE PURPLE CYAN DARK_RED DARK_GREEN DARK_BLUE GREY DARK_GREY LIGHT_YELLOW LIGHT_BLUE LIGHT_ORANGE ORANGE"
	);

	return(%data);
	}

sub parse_arguments
	{
	my %user_arguments = (
	    'SAMPLEINFO_FILE_FORMAT' => 'columns', # columns|rows
	    'OVERWRITE_FLAG' => 'FALSE',
	    'VERBOSE_FLAG' => 'TRUE',
	    'INTENSITY' => 'IDR', # IQR|IDR|<number>
	    'GENEINFO_IS_ORDERED_FLAG' => 'TRUE',
	    'SAMPLEINFO_IS_ORDERED_FLAG' => 'TRUE',
	    'CENTER_FLAG' => 'FALSE',
	    'PALETTE' => 'bright', # bright|muted
	    'HEATMAPS_COLOR' => 'GBR', # GBR|BWR
	    'VERSION_FLAG' => FALSE,
	    'HELP_FLAG' => FALSE
	    );

	GetOptions(
		'matrix=s' => \$user_arguments{'MATRIX_FILE'},
		'geneinfo=s' => \$user_arguments{'GENEINFO_FILE'},
		'sampleinfo=s' => \$user_arguments{'SAMPLEINFO_FILE'},
		'geneinfo-ordered=s' => \$user_arguments{'GENEINFO_IS_ORDERED_FLAG'},
		'sampleinfo-ordered=s' => \$user_arguments{'SAMPLEINFO_IS_ORDERED_FLAG'},
	    	'colors=s' => \$user_arguments{'COLORS_FILE'},
	    	'sampleinfo-format=s' => \$user_arguments{'SAMPLEINFO_FILE_FORMAT'},    	
	    	'output=s' => \$user_arguments{'PREFIX'},
	    	'overwrite=s' => \$user_arguments{'OVERWRITE_FLAG'},
	    	'verbose=s' => \$user_arguments{'VERBOSE_FLAG'},
	    	'intensity=s' => \$user_arguments{'INTENSITY'},
	    	'center=s' => \$user_arguments{'CENTER_FLAG'},
	    	'palette=s' => \$user_arguments{'PALETTE'},
	    	'heatmap-colors=s' => \$user_arguments{'HEATMAPS_COLOR'},
	    	'version' => \$user_arguments{'VERSION_FLAG'},
	    	'help' => \$user_arguments{'HELP_FLAG'}
        	)
	or die("[ERROR]\n");

	### show help if --help

	display_help() if ($user_arguments{'HELP_FLAG'} == TRUE);

	### show version if --version

	display_version() if ($user_arguments{'VERSION_FLAG'} == TRUE);

	### check if matrix file is defined and is readable

	die("[ERROR] Matrix file is required.") if (!defined($user_arguments{'MATRIX_FILE'}));

	die("[ERROR] File '$user_arguments{'MATRIX_FILE'}' not found.") if (!-e $user_arguments{'MATRIX_FILE'});

	### output prefix is set automatically if not user-defined

	$user_arguments{'PREFIX'} = $user_arguments{'MATRIX_FILE'}.".output"  if (!defined($user_arguments{'PREFIX'}));

	check_input_arguments(%user_arguments);

	

	$VERBOSE = TRUE if $user_arguments{'VERBOSE_FLAG'} eq "TRUE";

	return(%user_arguments);
	}

sub check_input_arguments
	{
	my %user_arguments = @_;
	
	die "[ERROR] File '$user_arguments{'PREFIX'}.xlsx' already exists." if ( (-e "$user_arguments{'PREFIX'}.xlsx") and ($user_arguments{'OVERWRITE_FLAG'} eq "FALSE") );
	die "[ERROR] Argument --palette must be [bright|muted].\n" if ( ($user_arguments{'PALETTE'} ne "bright") and  ($user_arguments{'PALETTE'} ne "muted") );
	die "[ERROR] Argument --center must be [TRUE|FALSE].\n" if ( ($user_arguments{'CENTER_FLAG'} ne "TRUE") and  ($user_arguments{'CENTER_FLAG'} ne "FALSE") );
	die "[ERROR] Argument --geneinfo-ordered must be [TRUE|FALSE].\n" if ( ($user_arguments{'GENEINFO_IS_ORDERED_FLAG'} ne "TRUE") and  ($user_arguments{'GENEINFO_IS_ORDERED_FLAG'} ne "FALSE") );
	die "[ERROR] Argument --sampleinfo-ordered must be [TRUE|FALSE].\n" if ( ($user_arguments{'SAMPLEINFO_IS_ORDERED_FLAG'} ne "TRUE") and  ($user_arguments{'SAMPLEINFO_IS_ORDERED_FLAG'} ne "FALSE") );
	die "[ERROR] Argument --overwrite must be [TRUE|FALSE].\n" if ( ($user_arguments{'OVERWRITE_FLAG'} ne "TRUE") and  ($user_arguments{'OVERWRITE_FLAG'} ne "FALSE") );
	die "[ERROR] Argument --verbose must be [TRUE|FALSE].\n" if ( ($user_arguments{'VERBOSE_FLAG'} ne "TRUE") and  ($user_arguments{'VERBOSE_FLAG'} ne "FALSE") );
	die "[ERROR] Argument --intensity must be [IQR|IDR|<positive number larger than 0>].\n" if ( ($user_arguments{'INTENSITY'} ne "IQR") and ($user_arguments{'INTENSITY'} ne "IDR") and (!looks_like_number($user_arguments{'INTENSITY'})) );
	die "[ERROR] Argument --intensity must be [IQR|IDR|<positive number larger than 0>].\n" if ( (looks_like_number($user_arguments{'INTENSITY'})) and ($user_arguments{'INTENSITY'} <= 0) );
	die "[ERROR] Argument --heatmap-colors must be [GBR|BWR].\n" if (  ($user_arguments{'HEATMAPS_COLOR'} ne "GBR") and  ($user_arguments{'HEATMAPS_COLOR'} ne "BWR") );
	die "[ERROR] Argument --sampleinfo-format must be {columns|rows}.\n" if (  ($user_arguments{'SAMPLEINFO_FILE_FORMAT'} ne "columns") and  ($user_arguments{'SAMPLEINFO_FILE_FORMAT'} ne "rows") );

	display_input_arguments(%user_arguments) if ($user_arguments{'VERBOSE_FLAG'} eq "TRUE");
	}

sub display_input_arguments
	{
	my %user_arguments = @_;

	print "
GETS (Gene Expression To Spreadsheet)
Version 1.0.0 (updated in November, 2019), Author: Daniel Aguilar\n//\n";

	print "[INPUT ARGUMENTS] Input data files:\n";
	print "[INPUT ARGUMENTS] \tExpression matrix file: '$user_arguments{'MATRIX_FILE'}'\n";
	print "[INPUT ARGUMENTS] \tGene info file: '$user_arguments{'GENEINFO_FILE'}'\n" if (defined($user_arguments{'GENEINFO_FILE'}));
	print "[INPUT ARGUMENTS] \tSample info file: '$user_arguments{'SAMPLEINFO_FILE'}'\n" if (defined($user_arguments{'SAMPLEINFO_FILE'}));
	print "[INPUT ARGUMENTS] \tAdditional information file: '$user_arguments{'MOREINFO_FILE'}'\n" if (defined($user_arguments{'MOREINFO_FILE'}));

	print "[INPUT ARGUMENTS] Colors and formats:\n";
	print "[INPUT ARGUMENTS] \tColors file: $user_arguments{'COLORS_FILE'}\n" if (defined($user_arguments{'COLORS_FILE'}));
	print "[INPUT ARGUMENTS] \tLimits for heatmap intensities: $user_arguments{'INTENSITY'}\n";
	print "[INPUT ARGUMENTS] \tHeatmap color palette: $user_arguments{'HEATMAPS_COLOR'}\n";
	print "[INPUT ARGUMENTS] \tGene/sample information color palette: $user_arguments{'PALETTE'}\n";

	print "[INPUT ARGUMENTS] Matrix pre-processing:\n";
	print "[INPUT ARGUMENTS] \tCenter expression: $user_arguments{'CENTER_FLAG'}\n";

	print "[INPUT ARGUMENTS] Gene information pre-processing:\n";
	print "[INPUT ARGUMENTS] \tGene names are ordered: $user_arguments{'GENEINFO_IS_ORDERED_FLAG'}\n";

	print "[INPUT ARGUMENTS] Sample information pre-processing:\n";
	print "[INPUT ARGUMENTS] \tSample info file format: $user_arguments{'SAMPLEINFO_FILE_FORMAT'}\n";
	print "[INPUT ARGUMENTS] \tSample names are ordered: $user_arguments{'SAMPLEINFO_IS_ORDERED_FLAG'}\n";

	print "[INPUT ARGUMENTS] Miscellaneous:\n";
	print "[INPUT ARGUMENTS] \tOutput prefix: '$user_arguments{'PREFIX'}'\n";
	print "[INPUT ARGUMENTS] \tOverwrite output file: $user_arguments{'OVERWRITE_FLAG'}\n";
	print "[INPUT ARGUMENTS] \tVerbose standard output: $user_arguments{'VERBOSE_FLAG'}\n";

	print "//\n";
	}

sub read_colors_file
    {
    my ($user_arguments_ref, $format_ref, $data_ref) = @_;

    my %color_data;

	if (!defined($$user_arguments_ref{'COLORS_FILE'}))
		{return(%color_data);}
	else
		{
		if (open(my $FHANDLE, '<', $$user_arguments_ref{'COLORS_FILE'}))
		      {
		      while (<$FHANDLE>)
				{
				chomp;

				next if ($_ !~ /^(GENEINFO|SAMPLEINFO)\t/);

				my ($section, $category, $value, $color) = split(/\t/);

				if ( ($category eq "") or ($color eq "") or ( ($value eq "") and ($color ne "GRADIENT_RED") and ($color ne "GRADIENT_GREEN") and ($color ne "GRADIENT_TWO_COLORS") and ($color ne "AUTO") ) )
					{
					error($data_ref, "In file '$$user_arguments_ref{'COLORS_FILE'}': Wrong format.\n");
					}
				elsif ( (!exists($$format_ref{$color})) and ($color !~ /GRADIENT/) and ($color ne "AUTO") )
					{
					error($data_ref, "In file '$$user_arguments_ref{'COLORS_FILE'}': Color '$color' is not understood.\n");
					}
				elsif ($color eq "AUTO")
					{
					$color_data{$section}{$category}{AUTO} = TRUE;
					}
				elsif ($color =~ /GRADIENT/)
					{
					$color_data{$section}{$category}{GRADIENT} = $color;
					}
				else
					{
					$color_data{$section}{$category}{COLOR}{$value} = $color;
					}

		          }
		      }
		  else
		      {error($data_ref, "In file '$$user_arguments_ref{'COLORS_FILE'}': cannot read file.\n");}
		}
  
    return(%color_data);
    }

sub order_geneinfo
	{
	my ($geneinfo_matrix_unordered_ref, $user_arguments_ref, $data_ref) = @_;

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{print "[GENE INFO] Ordering gene information.\n";}

	my %gene_row;

	### check that there are no duplicated row names in the matrix

	my %gene_names_hash;
	foreach my $gene (@{$$data_ref{GENE_NAMES_IN_MATRIX}})
		{
		$gene_names_hash{$gene} ++;

		if ($gene_names_hash{$gene} > 1)
			{error($data_ref, "In file '$$user_arguments_ref{'MATRIX_FILE'}': gene/probe name $gene is duplicated while --geneinfo-ordered=FALSE\n");}
		}

	for my $i (1..$#$geneinfo_matrix_unordered_ref)
		{
		my $gene = $$geneinfo_matrix_unordered_ref[$i][0];

		$gene_row{$gene} = $i if (!exists($gene_row{$gene})); ### if there're duplicated gene/probe names in the gene info file, only the first occurrence is considered
		}


	my @geneinfo_matrix_ordered;

	@{$geneinfo_matrix_ordered[0]} =  @{$$geneinfo_matrix_unordered_ref[0]}; ### The column names remain the same

	my $n = 0;
	for my $i (0..$#{$$data_ref{GENE_NAMES_IN_MATRIX}})
		{
		my $gene = ${$$data_ref{GENE_NAMES_IN_MATRIX}}[$i];

		if (exists($gene_row{$gene})) ### if --geneinfo-ordered=FALSE and there is a match between a gene/probe name in the gene info file and gene/probe name in the matrix
			{
			@{$geneinfo_matrix_ordered[$i+1]} = @{$$geneinfo_matrix_unordered_ref[$gene_row{$gene}]};
			}
		else 		### if --geneinfo-ordered=FALSE and a gene in matrix is NOT present in gene info file, its row will be filled with blank cells
			{
			$n ++;

			$geneinfo_matrix_ordered[$i+1][0] = "";

			for my $k (1..$#{$geneinfo_matrix_ordered[0]})
				{push(@{$geneinfo_matrix_ordered[$i+1]}, " ");}
			}
		}

	if ($n == (scalar(@geneinfo_matrix_ordered) - 1) )
		{error($data_ref, "Could not match any gene/probe name in gene information file gene/probe names in the matrix file.\n");}

	print "[GENE INFO] [WARNING] When re-ordering genes/probes: $n genes/probes were not present in the gene information file.\n" if ($n > 0);

	return(\@geneinfo_matrix_ordered);
	}



sub load_sampleinfo
	{
	my ($file, $format, $data_ref) = @_;

	my @sampleinfo_matrix;

	 ### if the sample info data is formatted as columns, we load it into sampleinfo_matrix as is
		
	my $nrow = 0;
	my $ncol_last;

	if (open(my $FHANDLE, '<', $file))
		{
		while (<$FHANDLE>)
				{
				chomp;

				next if (/^[!|#]/);
				next if ($_ !~ /\w/);

				@{$sampleinfo_matrix[$nrow]} = split(/\t/, $_);

				if ( (defined($ncol_last)) and ($ncol_last != scalar(@{$sampleinfo_matrix[$nrow]})) )
					{
					error($data_ref, "In file '$file': Variable number of columns.");
					}

				$ncol_last = scalar(@{$sampleinfo_matrix[$nrow]});

				$nrow ++;
				}
		}
	else
		{error($data_ref, "In file '$file': cannot read file.\n");}
		
	if ($format eq "rows")					### if the sample info data is vertically formatted ('rows'), we transpose it before loading it into sampleinfo_matrix
		{
		my @sampleinfo_matrix_transposed;

		for my $i (0..$#sampleinfo_matrix)
			{
			for my $j (0..$#{$sampleinfo_matrix[$i]})
				{
				$sampleinfo_matrix_transposed[$j][$i] = $sampleinfo_matrix[$i][$j];
				}
			}

		@sampleinfo_matrix = @sampleinfo_matrix_transposed;
		}


	### We make sure that all rows have the same number of columns, filling with blanks if necessary

	for my $i (0..$#sampleinfo_matrix)
		{
		for my $j (0..$#{$sampleinfo_matrix[$i]})
			{
			$sampleinfo_matrix[$i][$j] = "" if (!defined($sampleinfo_matrix[$i][$j]));
			}
		}

	return(\@sampleinfo_matrix);
	}


sub order_sampleinfo
	{
	my ($sampleinfo_matrix_unordered_ref, $user_arguments_ref, $data_ref) = @_;

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{print "[SAMPLE INFO] Ordering sample information.\n";}

	my $last_row = $#$sampleinfo_matrix_unordered_ref;

	my %sample_col;

	my %sample_names_hash;

	for my $i (1..$#{$$sampleinfo_matrix_unordered_ref[$last_row]})
		{
		my $sample = $$sampleinfo_matrix_unordered_ref[$last_row][$i];

		$sample_col{$sample} = $i; ### if there are dupicated sample names, only considers the info associated to the first occurence
		}

	%sample_names_hash = ();
	foreach my $sample (@{$$data_ref{SAMPLE_NAMES_IN_MATRIX}})
		{
		$sample_names_hash{$sample} ++;

		if ($sample_names_hash{$sample} > 1)
			{error($data_ref, "In file '$$user_arguments_ref{'MATRIX_FILE'}': sample name $sample is duplicated while --sampleinfo-ordered=FALSE\n");}
		}

	%sample_names_hash = ();

	my @sampleinfo_matrix_ordered;

	for my $i (0..$last_row)
		{
		$sampleinfo_matrix_ordered[$i][0] = $$sampleinfo_matrix_unordered_ref[$i][0]; ### The first column (row names) remain the same
		}


	my $n = 0;
	for my $i (0..$#{$$data_ref{SAMPLE_NAMES_IN_MATRIX}})
		{
		my $sample = ${$$data_ref{SAMPLE_NAMES_IN_MATRIX}}[$i];

		if (exists($sample_col{$sample}))
			{
			for my $k (0..$last_row)
				{$sampleinfo_matrix_ordered[$k][$i+1] = $$sampleinfo_matrix_unordered_ref[$k][$sample_col{$sample}];}
			}
		else 		### if --sampleinfo-ordered=FALSE and a sample in matrix is not present in sample info file, its column will be filled with blank cells
			{
			$n ++;

			for my $k (0..$last_row)
				{$sampleinfo_matrix_ordered[$k][$i+1] = " ";}
			}

		
		}

	if ($n == $$data_ref{'N_SAMPLES_MATRIX'})
		{error($data_ref, " When re-ordering samples: none of the samples of the matrix was present in the sample information file.\n");}

	print "[SAMPLE INFO] [WARNING] When re-ordering samples: $n samples of the matrix were not present in the sample information file.\n" if ($n > 0);

	print "//\n";

	return(\@sampleinfo_matrix_ordered);
	}

sub load_geneinfo
	{
	my ($file, $data_ref) = @_;

	my @matrix;

	my $n_row = 0;
	my $n_col = 0; ### the number of columns is determined by the number of column names

	if (open(my $FHANDLE, '<', $file))
		{
		while (<$FHANDLE>)
			{
			chomp;

			next if (/^[!|#]/);
			#next if ($_ !~ /\w/); ### We allow empty rows because it may be possible that a row in the gene info file has blank cells for all columns

			@{$matrix[$n_row]} = split(/\t/);

			if ($n_row == 0)
				{$n_col = $#{$matrix[$n_row]};}
			
			$n_row ++;
		    	}
       	}
	else
		{error($data_ref, "In file '$file': cannot read file.\n");}


	### We make sure that all rows have the same number of columns by filling empty cells with blanks

	for my $i (0..$#matrix)
		{
		for my $j (0..$n_col)
			{
			$matrix[$i][$j] = "" if (!defined($matrix[$i][$j]));
			}
		}


	return(\@matrix);
	}

sub process_geneinfo_data
	{
	my ($user_arguments_ref, $data_ref, $data_colors_ref, $worksheet, %format) = @_;

	if (!defined($$user_arguments_ref{'GENEINFO_FILE'}))
		{return(%$data_ref);}

	my $geneinfo_matrix_ref = load_geneinfo($$user_arguments_ref{'GENEINFO_FILE'}, $data_ref);

	if ($$user_arguments_ref{'GENEINFO_IS_ORDERED_FLAG'} eq "FALSE")
		{
		$geneinfo_matrix_ref = order_geneinfo($geneinfo_matrix_ref, $user_arguments_ref, $data_ref);
		}

	my @column_names = @{$$geneinfo_matrix_ref[0]};

	my %column_names_hash = map { $_ => TRUE } @column_names;

	my %colours_worksheet;

	if (exists($$data_colors_ref{GENEINFO})) ### if gene colors are defined
		{
		foreach my $category (keys(%{$$data_colors_ref{GENEINFO}}))
		      {
			next if (!exists($column_names_hash{$category}));

			if (exists($$data_colors_ref{GENEINFO}{$category}{GRADIENT}))
				{
		            $colours_worksheet{$category}{DATA_TYPE} = $$data_colors_ref{GENEINFO}{$category}{GRADIENT};

				my $j = first_index { $_ eq $category } @column_names;

				my @value_list;
				foreach my $i (1..$#$geneinfo_matrix_ref)
					{
					push(@value_list, $$geneinfo_matrix_ref[$i][$j]);
					}

				%{$colours_worksheet{$category}{CELL_TYPE}} = set_gradient(\@value_list, $colours_worksheet{$category}{DATA_TYPE}, $data_ref);
				}
			elsif (exists($$data_colors_ref{GENEINFO}{$category}{AUTO}))
				{
				my $j = first_index { $_ eq $category } @column_names;

				my @value_list;
				foreach my $i (1..$#$geneinfo_matrix_ref)
					{
					push(@value_list, $$geneinfo_matrix_ref[$i][$j]);
					}

				$colours_worksheet{$category}{DATA_TYPE} = "AUTO";
				%{$colours_worksheet{$category}{CELL_TYPE}} = autocolors($$data_ref{COLOR_LIST}, \@value_list);
				}
			else
				{
				$colours_worksheet{$category}{DATA_TYPE} = "user-defined";

				foreach my $value (keys(%{$$data_colors_ref{GENEINFO}{$category}{COLOR}}))
				    {
				    $colours_worksheet{$category}{CELL_TYPE}{$value} = $$data_colors_ref{GENEINFO}{$category}{COLOR}{$value};
				    }  
				}             
		      }
		}

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[GENE INFO] Number of rows: ".scalar(@$geneinfo_matrix_ref)."\n";
		print "[GENE INFO] Number of columns: ".scalar(@{$$geneinfo_matrix_ref[0]})."\n";

		if (scalar(keys(%colours_worksheet)) > 0)
			{
			my $geneinfo_hash_ref = geneinfo_matrix_to_hash($geneinfo_matrix_ref);

			print "[GENE INFO] User-defined colors:\n";

			foreach my $category (sort(keys(%colours_worksheet)))
				{
				print "[GENE INFO] \t$category: $colours_worksheet{$category}{DATA_TYPE}\n";
				}
			}

		print "//\n";

		}

	### gene info data starts from $$data_ref{'SAMPLE_INFO_ROWS'} + 1 and $$data_ref{'MATRIX_DATA_COLUMNS'} + 1

	my %col_size;
	my $min = 5;
	my $max = 50;

	for my $j (0..$#{$$geneinfo_matrix_ref[0]})
		{
		my $col = $$data_ref{'MATRIX_DATA_COLUMNS'} + 1 +$j;

		$$data_ref{'GENE_INFO_COLUMN_WIDTH'}{$col} = $min; # default
		}

	for my $i (1..$#$geneinfo_matrix_ref)
		{
		for my $j (0..$#{$$geneinfo_matrix_ref[$i]})
			{
			my $col = $$data_ref{'MATRIX_DATA_COLUMNS'} + 1 +$j;

			my $length = length($$geneinfo_matrix_ref[$i][$j]) * 2;
			$length = $max if ($length > $max);

			$$data_ref{'GENE_INFO_COLUMN_WIDTH'}{$col} = $length if ($length > $$data_ref{'GENE_INFO_COLUMN_WIDTH'}{$col});
			}
		}

	for my $i (0..$#$geneinfo_matrix_ref)
		{
		if ($i == 0) ### we are in the first row of the GENEINFO_FILE, which contains the headers
			{
			for my $j (0..$#{$$geneinfo_matrix_ref[$i]})
				{
				my $col = $$data_ref{'MATRIX_DATA_COLUMNS'} + 1 +$j;

                		$worksheet->write($$data_ref{'SAMPLE_INFO_ROWS'} + 1 + $i, $$data_ref{'MATRIX_DATA_COLUMNS'} + 1 +$j, $$geneinfo_matrix_ref[$i][$j], $format{GENEINFO_LABELS}); # print the headers in GENEINFO_LABELS cell format
				}
			}
		else
			{
			for my $j (0..$#{$$geneinfo_matrix_ref[$i]})
				{
 				my $f = "GENEINFO";  ### the cell format will be GENEINFO unless it is defined otherwise in GENEINFO_COLORS_FILE

                   	 if (exists($colours_worksheet{$column_names[$j]}{CELL_TYPE}{$$geneinfo_matrix_ref[$i][$j]}))
				        {$f = $colours_worksheet{$column_names[$j]}{CELL_TYPE}{$$geneinfo_matrix_ref[$i][$j]};}

				$worksheet->write($$data_ref{'SAMPLE_INFO_ROWS'} + 1 + $i, $$data_ref{'MATRIX_DATA_COLUMNS'} + 1 +$j, $$geneinfo_matrix_ref[$i][$j], $format{$f});
				}
			}
		}

	if (scalar(@$geneinfo_matrix_ref) != $$data_ref{MATRIX_DATA_ROWS})
		{error($data_ref, "In file '$$user_arguments_ref{'GENEINFO_FILE'}': the number of rows (".scalar(@$geneinfo_matrix_ref).") does not match the number of rows in '$$user_arguments_ref{'MATRIX_FILE'}' ($$data_ref{MATRIX_DATA_ROWS})\n");}

	$$data_ref{'GENE_INFO_COLUMNS'} = scalar(@column_names);

	return(%$data_ref);
	}

sub geneinfo_matrix_to_hash
	{
	my $geneinfo_matrix_ref = $_[0];

	my @column_names = @{$$geneinfo_matrix_ref[0]};

	my %hash;

	for my $i (0..$#$geneinfo_matrix_ref)
		{
		next if ($i == 0);

		for my $j (0..$#{$$geneinfo_matrix_ref[$i]})
			{
			$hash{$column_names[$j]}{$$geneinfo_matrix_ref[$i][$j]} = TRUE;
			}
		}

	return(\%hash);
	}

sub format_output_file
	{
	my ($worksheet, $data_ref) = @_;

	$worksheet->set_column(0, 0, 12); ### width 12 for column 0
	$worksheet->set_column(1, $$data_ref{'MATRIX_DATA_COLUMNS'}, 3);   ### width 3 for all matrix columns
	$worksheet->set_row($$data_ref{'SAMPLE_NAMES_ROW'}, $$data_ref{'SAMPLE_NAMES_ROW_HEIGHT'});

	foreach my $col (sort(keys(%{$$data_ref{'GENE_INFO_COLUMN_WIDTH'}})))
		{
		$worksheet->set_column($col, $col, $$data_ref{'GENE_INFO_COLUMN_WIDTH'}{$col} );    ### width 20 for all geneinfo columns
		}

	
	}

sub process_matrix_data
	{
	my ($user_arguments_ref, $data_ref, $worksheet, $matrix_ref, %format) = @_;

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
        	print "[MATRIX] Processing matrix data...\n";
		print "[MATRIX] Dimensions: ".(scalar(@$matrix_ref))." x ".(scalar(@{$$matrix_ref[0]}))."\n";
		print "[MATRIX] Number of samples: ".(scalar(@{$$matrix_ref[0]}) - 1)."\n";
		print "[MATRIX] Number of genes: ".(scalar(@$matrix_ref) - 1)."\n";
		}

	my @intensity_range = calc_intensity_range($matrix_ref, $user_arguments_ref, $data_ref);
	
	my $row = $$data_ref{'SAMPLE_INFO_ROWS'} + 1; ### matrix data starts from this row + 1

	$$data_ref{'SAMPLE_NAMES_ROW_HEIGHT'} = 30;

	for my $mrow (0..$#$matrix_ref)
		{
		if ($mrow == 0) ### the first row of $matrix_ref contains the sample names
			{
			my $length = length(${$$matrix_ref[$mrow]}[0]) * 10;

			$$data_ref{'SAMPLE_NAMES_ROW_HEIGHT'} = $length if ($$data_ref{'SAMPLE_NAMES_ROW_HEIGHT'} < $length);

			$worksheet->write($row, 0, ${$$matrix_ref[$mrow]}[0], $format{GENEINFO});

			for my $col (1..$#{$$matrix_ref[$mrow]})
				{
				$$data_ref{'SAMPLE_NAMES_ROW'} = $row;

                		$worksheet->write($row, $col, ${$$matrix_ref[$mrow]}[$col], $format{SAMPLE_IDS});
                		}
			}
		else  ### the remaining rows of $matrix_ref contain the expression values and are coloured according to intensity gradient
		      {
		      $worksheet->write($row, 0, ${$$matrix_ref[$mrow]}[0], $format{GENEINFO});

			push(@{$$data_ref{'GENE_NAMES_IN_MATRIX'}}, ${$$matrix_ref[$mrow]}[0]);

			for my $col (1..$#{$$matrix_ref[$mrow]})
				 {
				my $value = ${$$matrix_ref[$mrow]}[$col];

			
				    my $format;

				    if ($value < $intensity_range[10])
					    {
					    for my $i (0..9)
						    {
						    my $label = "gradient $i";

						    if (abs($value) > abs($intensity_range[$i]))
							    {
							    error($data_ref, "[ERROR] No format '$label'") if (!exists($format{$label}));

							    $format = $format{$label};

							    last;
							    }
						    }

					   
						}
					elsif ($value > $intensity_range[10])
					    {
					    for(my $i = 20; $i >= 11; $i --)
						    {
						    my $label = "gradient $i";

						    if ($value > $intensity_range[$i])
							    {
							    error($data_ref, "[ERROR] No format '$label'") if (!exists($format{$label}));

							    $format = $format{$label};
							    last;
							    }
						    }
					    }

				    if (!defined($format))
					    {
					    $format = $format{'gradient 10'};
					    }
				    
				$worksheet->write($row, $col, sprintf("%.3f", $value), $format);
				}
		      }

		$row ++;
		}

	
	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[MATRIX] Heatmap intensity range limits have been (automatically) set to: $intensity_range[0] to $intensity_range[$#intensity_range]";

		print ", centered around 0" if ($intensity_range[10] == 0);

		print ".\n";
		print "//\n";
		}

	$$data_ref{'MATRIX_DATA_ROWS'} = scalar(@$matrix_ref);
	$$data_ref{'MATRIX_DATA_COLUMNS'} = scalar(@{$$matrix_ref[0]}) - 1; ### because the worksheet starts from (0,0)
	$$data_ref{'SPREADHSEET_ROWS'} = $row - 1;
	

	return(%$data_ref);
	}


sub error
	{
	my ($data_ref, $message) = @_;
	
	print "[ERROR] $message\n";

	system("rm", "-f", $$data_ref{'SPREADSHEET_FILE'}) if (-e $$data_ref{'SPREADSHEET_FILE'});

	die();
	}

sub warning
	{
	my ($data_ref, $message) = @_;

	$message =~ s/\]/\] Warning:/;
	
	print "$message\n" if $VERBOSE == TRUE;
	}



sub process_sampleinfo
	{
	my ($user_arguments_ref, $data_ref, $worksheet, $matrix_ref, $data_colors_ref, %format) = @_;

	if (!defined($$user_arguments_ref{'SAMPLEINFO_FILE'}))
		{return(%$data_ref);}

	my $n_samples_matrix = scalar(@{$$matrix_ref[0]}) - 1; ### -1 because the first column contains gene names

	@{$$data_ref{'SAMPLE_NAMES_IN_MATRIX'}} = (@{$$matrix_ref[0]})[1..$n_samples_matrix];


	my $sampleinfo_matrix_ref = load_sampleinfo($$user_arguments_ref{'SAMPLEINFO_FILE'}, $$user_arguments_ref{'SAMPLEINFO_FILE_FORMAT'}, $data_ref);

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		my $n_columns_sampleinfo_matrix = scalar(@$sampleinfo_matrix_ref );
		my $n_rows_sampleinfo_matrix = scalar(@{$$sampleinfo_matrix_ref[0]});

		print "[SAMPLE INFO] Processing sample info data...\n";

		print "[SAMPLE INFO] Number of samples: ".($n_rows_sampleinfo_matrix-1)."\n"; ### -1 because the first column contains the category names
		print "[SAMPLE INFO] Number of information rows: $n_columns_sampleinfo_matrix\n";
		}


	if ($$user_arguments_ref{'SAMPLEINFO_IS_ORDERED_FLAG'} eq "FALSE")
		{
		$sampleinfo_matrix_ref = order_sampleinfo($sampleinfo_matrix_ref, $user_arguments_ref, $data_ref);
		}


	my $n_columns_sampleinfo_matrix = scalar(@$sampleinfo_matrix_ref );
	my $n_rows_sampleinfo_matrix = scalar(@{$$sampleinfo_matrix_ref[0]});

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[SAMPLE INFO] Processing sample info data...\n";

		print "[SAMPLE INFO] Number of samples: ".($n_rows_sampleinfo_matrix-1)."\n"; ### -1 because the first column contains the category names
		print "[SAMPLE INFO] Number of information rows: $n_columns_sampleinfo_matrix\n";
		}

	
	if ( ($n_rows_sampleinfo_matrix-1) != $n_samples_matrix)
		{
		error($data_ref, "In file '$$user_arguments_ref{'SAMPLEINFO_FILE'}': number of samples is ".($n_rows_sampleinfo_matrix-1)." but the matrix file has $n_samples_matrix samples.");
		}



	### the category names are stored in %category_name_hash

	my %category_name_hash;
	foreach my $i (0..$#$sampleinfo_matrix_ref)
		{
		$category_name_hash{$$sampleinfo_matrix_ref[$i][0]} = TRUE;
		}

	### Colour information

	my %colours_worksheet;	

	### we initialize all colors to black font over white background by default

	for my $i (0..$#$sampleinfo_matrix_ref)
		{
		my $category = $$sampleinfo_matrix_ref[$i][0];

		$colours_worksheet{$category}{DATA_TYPE} = "DISCRETE";

		for my $j (1..$#{$$sampleinfo_matrix_ref[$i]})
			{
			$colours_worksheet{$category}{COLOR}{$$sampleinfo_matrix_ref[$i][$j]} = "default";
			$colours_worksheet{$category}{CELL_TYPE}{$$sampleinfo_matrix_ref[$i][$j]} = "GENEINFO";
			}
		}

	if (!exists($$data_colors_ref{SAMPLEINFO})) ### if sample colors are not defined
		{
		print "[SAMPLE INFO] Sample info will have no color.\n" if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE");
		}
	else                                         ### if sample colors are defined
		{
		foreach my $category (keys(%{$$data_colors_ref{SAMPLEINFO}}))
		      {
			next if (!exists($category_name_hash{$category}));

			if (exists($$data_colors_ref{SAMPLEINFO}{$category}{GRADIENT}))
					{
					$colours_worksheet{$category}{DATA_TYPE} = $$data_colors_ref{SAMPLEINFO}{$category}{GRADIENT};

					 my @value_list = keys(%{$colours_worksheet{$category}{COLOR}});

					%{$colours_worksheet{$category}{CELL_TYPE}} = set_gradient(\@value_list, $colours_worksheet{$category}{DATA_TYPE}, $data_ref);
					}
			elsif (exists($$data_colors_ref{SAMPLEINFO}{$category}{AUTO}))
				      {
				       my @value_list = keys(%{$colours_worksheet{$category}{COLOR}});

				       $colours_worksheet{$category}{DATA_TYPE} = "AUTO";
				       %{$colours_worksheet{$category}{CELL_TYPE}} = autocolors($$data_ref{COLOR_LIST}, \@value_list);
				       }
			else
				{
				foreach my $value (keys(%{$$data_colors_ref{SAMPLEINFO}{$category}{COLOR}}))
					{
				    	$colours_worksheet{$category}{COLOR}{$value} = "user-defined";
		              	$colours_worksheet{$category}{DATA_TYPE} = "DISCRETE";
		              	$colours_worksheet{$category}{CELL_TYPE}{$value} = $$data_colors_ref{SAMPLEINFO}{$category}{COLOR}{$value};
		              	}
		          }

			}

     

		### stdout color info

		if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
			{
			print "[SAMPLE INFO] User-defined colors:\n";

			foreach my $category (sort(keys(%category_name_hash)))
				{
				print "[SAMPLE INFO] \t$category: ";

				if (exists($colours_worksheet{$category}{DATA_TYPE}))
					{
					print "$colours_worksheet{$category}{DATA_TYPE}\n";
					}
				else
					{
					print "not-defined.\n";
					next;
					}

				}
			}
		}


	### write spreadheet
    
	my $row = 0;    ## # the n of rows in sample info is added to %data
	for my $i (0..$#$sampleinfo_matrix_ref)
		{
		my $category = $$sampleinfo_matrix_ref[$i][0];

		$worksheet->write($row, 0, $category, $format{GENEINFO_LABELS}); ### category names (column 0 in the spreadheet)

		for my $j (1..$#{$$sampleinfo_matrix_ref[$i]})
			{
 			die "No color for '$category' value '$$sampleinfo_matrix_ref[$i][$j]'\n" if (!exists($colours_worksheet{$category}{CELL_TYPE}{$$sampleinfo_matrix_ref[$i][$j]}));
        
			die "No format >$colours_worksheet{$category}{CELL_TYPE}{$$sampleinfo_matrix_ref[$i][$j]}<\n" if (!exists($format{$colours_worksheet{$category}{CELL_TYPE}{$$sampleinfo_matrix_ref[$i][$j]}}));

			$worksheet->write($row, $j, $$sampleinfo_matrix_ref[$i][$j], $format{$colours_worksheet{$category}{CELL_TYPE}{$$sampleinfo_matrix_ref[$i][$j]}});
			}

        	$row ++;
		}

	$$data_ref{'SAMPLE_INFO_ROWS'} = scalar(@$sampleinfo_matrix_ref) - 1;## # -1 because $worksheet starts at position (0,0)

	print "//\n";

	return(%$data_ref);
	}


sub set_gradient
	{
	my ($values_list_ref, $type, $data_ref) = @_;

	my %value_hash;

	my %numeric_value_hash;

	foreach my $value (@$values_list_ref) 
		{
		if (!looks_like_number($value))
			{
			$value_hash{$value} = "GREY"; ### GREY color to non-numeric values
			}
		else
			{$numeric_value_hash{$value} = TRUE;}
		}

	my @numeric_value_list = keys(%numeric_value_hash);

	my $min = min(@numeric_value_list);
	my $max = max(@numeric_value_list);

	if ( ($type =~ /TWO_COLORS/) and ( ($min * $max) > 0) ) ### GRADIENT_TWO_COLORS is only allowed is the value range includes both positive and negative values, in which case will use 0 as centerpoint for the gradient
		{error($data_ref, "[ERROR] Color '$type' can only be used in the color range includes positive and negative values.");}

	if ($type !~ /TWO_COLORS/) ### If GRADIENT_RED or GRADIENT_GREEN is provided
		{
		my $range = (abs($max) + abs($min)) / 5;

		for my $j (0..$#numeric_value_list)
			{
			my $value = $numeric_value_list[$j];

			for my $k (0..5)
				{
				if ($type eq "GRADIENT_RED")
					{
					if ( ($value >= ($min+($range*$k))) and ($value < ($min+($range*($k+1)))) )
						{
						if ($k == 0)
							{$value_hash{$value} = "WHITE";}
						else
							{$value_hash{$value} = "SCALE_".$k."_RED";}

						last;
						}
					}
				elsif ($type eq "GRADIENT_GREEN")
					{
					if ( ($value >= ($min+($range*$k))) and ($value < ($min+($range*($k+1)))) )
						{
						if ($k == 0)
							{$value_hash{$value} = "WHITE";}
						else
							{$value_hash{$value} = "SCALE_".$k."_GREEN";}

						last;
						}
					}
				else
					{error($data_ref, "[ERROR] Undefined error '$type'.");}
				}
			}
		}
	else		### If GRADIENT_TWO_COLORS is provided
		{
		my $range_pos = abs($max) / 5;
		my $range_neg = abs($min) / 5;

		for my $j (0..$#numeric_value_list)
			{
			my $value = $numeric_value_list[$j];

			for my $k (0..5)
				{
				if ( ($value >= ($min+($range_neg*$k))) and ($value < ($min+($range_neg*($k+1)))) and ($value < 0) ) 
					{
					my $color = "SCALE_0";

					$color = "SCALE_".(5-$k)."_GREEN" if ($k < 5);

					$value_hash{$value} = $color;
					last;
					}
				elsif ( ($value >= (0+($range_pos*$k))) and ($value < (0+($range_pos*($k+1)))) and ($value > 0) ) 
					{
					my $color = "SCALE_0";

					$color = "SCALE_".$k."_RED" if ($k > 0);

					$value_hash{$value} = $color;
					last;
					}
				}
			}
		}

	

	return(%value_hash);
	}



sub define_formats
	{
	my ($workbook, $user_arguments_ref, $data_ref, %color) = @_;

	my %format;

	my @format_list = ("NO_FORMAT", "SAMPLE_IDS", "GENEINFO", "GENEINFO_LABELS");

	push(@format_list, ("WHITE", "MAGENTA", "LIME", "RED", "GREEN", "YELLOW", "BLUE", "PURPLE", "CYAN", "DARK_RED", "DARK_GREEN", "DARK_BLUE", "GREY", "DARK_GREY", "LIGHT_YELLOW", "LIGHT_BLUE", "LIGHT_ORANGE", "ORANGE"));
	
	### formats for the text cells (sample info and gene info)

	foreach my $key (@format_list)
		{
		$format{$key} = $workbook->add_format();
		$format{$key}->set_bg_color($color{WHITE});
		$format{$key}->set_color($color{BLACK});
		$format{$key}->set_size(9);
		$format{$key}->set_border(1);
		$format{$key}->set_align('center');

		if ($key eq "GENEINFO_LABELS")
			{
			$format{$key}->set_bold();
			}
		elsif ($key eq "GENEINFO")
			{
			$format{$key}->set_bold(0);
			}
		elsif ($key eq "SAMPLE_IDS")
			{
			$format{$key}->set_align('right');
			$format{$key}->set_rotation(90);
			$format{$key}->set_bold();
			}
		elsif ($key eq "WHITE")
			{$format{$key}->set_bg_color($color{WHITE});}
		elsif (defined($color{$key}))
			{$format{$key}->set_bg_color($color{$key});}
		elsif ($key ne "NO_FORMAT")
			{error($data_ref, "[ERROR] Color '$key' not defined.");}
		}



	### formats for the sample info cells with a gradient of values. Restricted to RED shades because of the limitation in the number of colors in MS Excel.


	foreach my $key ("SCALE_0", "SCALE_1", "SCALE_2", "SCALE_3", "SCALE_4", "SCALE_5")
		{
		if ($key eq "SCALE_0")
			{
			error($data_ref, "[ERROR] Color '$key' not defined.") if (!defined($color{$key}));

			$format{$key} = $workbook->add_format();
			$format{$key}->set_color($color{BLACK});
			$format{$key}->set_size(9);
			$format{$key}->set_border(1);
			$format{$key}->set_align('center');
			$format{$key}->set_bg_color($color{$key});

			next;
			}

		my $key2 = $key."_RED";

		error($data_ref, "[ERROR] Color '$key2' not defined.") if (!defined($color{$key2}));

		$format{$key2} = $workbook->add_format();
		$format{$key2}->set_color($color{BLACK});
		$format{$key2}->set_size(9);
		$format{$key2}->set_border(1);
		$format{$key2}->set_align('center');
		$format{$key2}->set_bg_color($color{$key2});

		$key2 = $key."_GREEN";

		error($data_ref, "[ERROR] Color '$key2' not defined.") if (!defined($color{$key2}));

		$format{$key2} = $workbook->add_format();
		$format{$key2}->set_color($color{BLACK});
		$format{$key2}->set_size(9);
		$format{$key2}->set_border(1);
		$format{$key2}->set_align('center');
		$format{$key2}->set_bg_color($color{$key2});
		}

	### formats for the gradients

	for my $i (0..20)
		{
		my $label = "gradient bk $i";
		my $label2 = "gradient $i";

		error($data_ref, "[ERROR] No color '$label2'.") if (!exists($color{$label2}));

		$format{$label} = $workbook->add_format();
		$format{$label}->set_size(9);
		$format{$label}->set_border(1);
		$format{$label}->set_align('center');
		$format{$label}->set_bg_color($color{$label2});
		$format{$label}->set_color($color{BLACK});
		}

	### formats for the heatmap cells

	for my $i (0..20)
		{
		my $label = "gradient $i";

		error($data_ref, "[ERROR] No color '$label'.") if (!exists($color{$label}));

		$format{$label} = $workbook->add_format();
		$format{$label}->set_size(9);
		$format{$label}->set_border(1);
		$format{$label}->set_align('center');
		$format{$label}->set_bg_color($color{$label});
		$format{$label}->set_color($color{$label});
		}



	return(%format)
	}


sub define_colors
	{
	my ($workbook, $user_arguments_ref, $data_ref) = @_;

	my %color;

	### colors 8-16, 18-23 and 54-56 are reserved for the gene information section

	if ($$user_arguments_ref{'PALETTE'} eq "default")
		{
		$color{RED} = $workbook->set_custom_color(8, 255, 0, 0);
		$color{GREEN} = $workbook->set_custom_color(9, 0, 255, 0);
		$color{BLUE} =  $workbook->set_custom_color(10, 0, 0, 255);
		$color{YELLOW} = $workbook->set_custom_color(11, 255, 255, 0);
		$color{PURPLE} = $workbook->set_custom_color(12, 255, 0, 127);
		$color{CYAN} = $workbook->set_custom_color(13, 0, 255, 255);
		$color{DARK_RED} = $workbook->set_custom_color(14, 153, 0, 0);
		$color{DARK_GREEN} = $workbook->set_custom_color(15, 0, 153, 0);
		$color{DARK_BLUE} = $workbook->set_custom_color(16, 0, 0, 153);
		$color{GREY} = $workbook->set_custom_color(18, 192, 192, 192);
		$color{DARK_GREY} = $workbook->set_custom_color(19, 64, 64, 64);
		$color{LIGHT_YELLOW} = $workbook->set_custom_color(20, 255, 255, 204);
		$color{LIGHT_BLUE} = $workbook->set_custom_color(21, 135,206,250);
		$color{LIGHT_ORANGE} = $workbook->set_custom_color(22, 255, 210, 50);
		$color{ORANGE} = $workbook->set_custom_color(23, 255,140,0);	
		$color{MAGENTA} = $workbook->set_custom_color(54, 255, 153, 255);
		$color{LIME} = $workbook->set_custom_color(55, 191, 255, 0);	
		$color{WHITE} = $workbook->set_custom_color(56, 255, 255, 255);		
		}
	else
		{
		$color{RED} = $workbook->set_custom_color(8, 255, 100, 100);
		$color{GREEN} = $workbook->set_custom_color(9, 100, 255, 100);
		$color{BLUE} =  $workbook->set_custom_color(10, 100, 100, 255);
		$color{YELLOW} = $workbook->set_custom_color(11, 255, 255, 100);
		$color{PURPLE} = $workbook->set_custom_color(12, 255, 100, 127);
		$color{CYAN} = $workbook->set_custom_color(13, 100, 255, 255);
		$color{DARK_RED} = $workbook->set_custom_color(14, 153, 50, 50);
		$color{DARK_GREEN} = $workbook->set_custom_color(15, 50, 153, 50);
		$color{DARK_BLUE} = $workbook->set_custom_color(16, 50, 50, 153);
		$color{GREY} = $workbook->set_custom_color(18, 127, 127, 127);
		$color{DARK_GREY} = $workbook->set_custom_color(19, 192, 192, 192);
		$color{LIGHT_YELLOW} = $workbook->set_custom_color(20, 255, 255, 204);
		$color{LIGHT_BLUE} = $workbook->set_custom_color(21, 165,246,255);
		$color{LIGHT_ORANGE} = $workbook->set_custom_color(22, 255, 210, 50);
		$color{ORANGE} = $workbook->set_custom_color(23, 255,140,0);	
		$color{MAGENTA} = $workbook->set_custom_color(54, 255, 227, 255);
		$color{LIME} = $workbook->set_custom_color(55, 230, 255, 158);		
		$color{WHITE} = $workbook->set_custom_color(56, 255, 255, 255);		
		}

	### colors 24-44 are reserved for the heatmap section

	my @color_begin;
	my @color_end;

	if ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "GBR")
		{
		@color_begin = (0, 0, 0);
		@color_end = (0, 255, 0);
		}
	elsif ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "BWR")
		{
		@color_begin = (255, 255, 255);
		@color_end = (0, 0, 255);
		}
	else
		{error($data_ref, "[ERROR]");}

	my $j = 0;

	my $label = "gradient $j";

	$color{$label} = $workbook->set_custom_color(24 + $j, $color_end[0], $color_end[1], $color_end[2]);

	$j ++;

	my @list = array_gradient(\@color_begin, \@color_end, 9);

	for my $i (0..$#list)
		{
		my $label = "gradient $j";

		my $r = int($list[$#list-$i][0]);
		my $g = int($list[$#list-$i][1]);
		my $b = int($list[$#list-$i][2]);

		$color{$label} = $workbook->set_custom_color(24 + $j, $r, $g, $b);

		$j ++;
		}


	$label = "gradient 10";

	if ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "GBR")
		{
		$color{$label} = $workbook->set_custom_color(24 + $j, 0, 0, 0);
		}
	elsif ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "BWR")
		{
		$color{$label} = $workbook->set_custom_color(24 + $j, 255, 255, 255);
		}

	$j ++;

	if ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "GBR")
		{
		@color_begin = (0, 0, 0);
		@color_end = (255, 0, 0);
		}
	elsif ($$user_arguments_ref{'HEATMAPS_COLOR'} eq "BWR")
		{
		@color_begin = (255, 255, 255);
		@color_end = (255, 0, 0);
		}
	
	@list = array_gradient(\@color_begin, \@color_end, 9);

	for my $i (0..$#list)
			{
			my $label = "gradient $j";

			my $r = int($list[$i][0]);
			my $g = int($list[$i][1]);
			my $b = int($list[$i][2]);

			$color{$label} = $workbook->set_custom_color(24 + $j, $r, $g, $b);

			$j ++;
			}

	$label = "gradient $j";

	$color{$label} = $workbook->set_custom_color(24 + $j, $color_end[0], $color_end[1], $color_end[2]);

	### colors 45 - 53 are reserved for the red gradient of the gradients of the gene/sample information

	$color{SCALE_0} = $color{WHITE};

	$color{SCALE_5_RED} = $workbook->set_custom_color(17, 255,0,0);
	$color{SCALE_4_RED} = $workbook->set_custom_color(49, 255,51,51);
	$color{SCALE_3_RED} = $workbook->set_custom_color(48, 255,102,102);
	$color{SCALE_2_RED} = $workbook->set_custom_color(47, 255,153,153);
	$color{SCALE_1_RED} = $workbook->set_custom_color(46, 255,204,204);

	$color{SCALE_5_GREEN} = $workbook->set_custom_color(45, 0,255,0);
	$color{SCALE_4_GREEN} = $workbook->set_custom_color(53, 51,255,51);
	$color{SCALE_3_GREEN} = $workbook->set_custom_color(52, 102,255,102);
	$color{SCALE_2_GREEN} = $workbook->set_custom_color(51, 153,255,153);
	$color{SCALE_1_GREEN} = $workbook->set_custom_color(50, 204,255,204);

	return(%color)
	}

sub load_and_center_matrix
	{
	my ($user_arguments_ref, $data_ref) = @_;

	### checking if the 1st row contains sample names and the other rows contain expression values
	### checking if the expression values are centered

	my @matrix;
	my $n_row = 0;
	my $n_column = 0;
	my $n_numbers_first_row = 0;
	my $n_numbers_first_column = 0;
	my $n_negative_values = 0;
	my $j = 0;

    if (open(my $FHANDLE, '<', $$user_arguments_ref{'MATRIX_FILE'}))
        {
        while (<$FHANDLE>)
            {
		chomp;

		next if (/^[!|#]/);
		next if ($_ !~ /\w/);

		my @cell_contents_list = split(/\t/, $_);

            @{$matrix[$n_row]} = @cell_contents_list;

            if ( ($n_row > 0) and (scalar(@{$matrix[$n_row]}) != $j) )
                {
                error($data_ref, "[MATRIX PROCESSING] In matrix file: not all rows have the same number of columns.");
                }

            $j = scalar(@{$matrix[$n_row]});

            if ($n_row == 0) ### first row: sample names
			{
			$matrix[$n_row][0] = "unnamed cell" if ( (!defined($matrix[$n_row][0])) or ($matrix[$n_row][0] !~ /\w/) );

			for my $i (1..$#cell_contents_list)
				{
				if (looks_like_number($cell_contents_list[$i]))
                        	{$n_numbers_first_row ++;}
				}
                }
            else         ### remaining rows: gene names + expression values
                {
                if (looks_like_number($cell_contents_list[0]))
				{$n_numbers_first_column ++;}

			$n_column ++;

			my $col = 1;
			for my $i (1..$#cell_contents_list)
				{
				my $value = $cell_contents_list[$i];

				if (!looks_like_number($value))
					{
		                  error($data_ref, "[MATRIX PROCESSING] In matrix file: value in column $col, row ".($n_row+1)." ('$value') does not look like a number.\n");
					}
				elsif ($value < 0)
					{$n_negative_values ++;}

				$col ++;
				}
                }

		$n_row ++;
            }
        }
    else
        {error($data_ref, "[MATRIX] In matrix file: cannot read file.\n");}



	if ($j <= 1)
		{
		error($data_ref, "[MATRIX PROCESSING] Matrix must have >= 2 columns.");
		}
	if ($n_row <= 1)
		{
		error($data_ref, "[MATRIX PROCESSING] Matrix must have >= 2 rows.");
		}

	if ($n_numbers_first_row == $n_row)
		{
		warning($data_ref, "[MATRIX PROCESSING] In matrix file: all sample names look like numbers.");
		}

	if ($n_numbers_first_column == $n_column)
		{
		warning($data_ref, "[MATRIX PROCESSING] In matrix file: all gene names look like numbers.");
		}

	$$data_ref{'N_GENES_MATRIX'} = scalar(@matrix);
	$$data_ref{'N_SAMPLES_MATRIX'} = scalar(@{$matrix[0]}) - 1;


    if ($$user_arguments_ref{'CENTER_FLAG'} eq 'FALSE')
        {
       

        return(@matrix);
        }

	for my $i (1..$#matrix)
		{
		my $stat = Statistics::Descriptive::Full->new();
		$stat->add_data(@{$matrix[$i]}[1 .. ($j-1)]);

		my $median = $stat->percentile(50);
			
		for my $j (1..$#{$matrix[$i]})
			{
			$matrix[$i][$j] -= $median;
			}
		}

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[MATRIX PROCESSING] Values in the matrix have been centered.\n";
		print "//\n";
		}	
	

	return(@matrix);
	}


sub calc_intensity_range
	{
	my ($matrix_ref, $user_arguments_ref, $data_ref) = @_;

	my $nrows = scalar(@$matrix_ref) - 1;
	my $ncols = scalar(@{$$matrix_ref[0]}) - 1;


	my $max_n_cells = $ncols * $nrows;

	$max_n_cells = SAMPLE_MATRIX if ($max_n_cells > SAMPLE_MATRIX);

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[MATRIX] Sampling $max_n_cells random values to calculate the intensity range.\n";
		}

	my $all_positive_flag = TRUE;
	my $all_negative_flag = TRUE;

	my @total_intensities_list;

	my %duplicated;

	my @sample_list;

	for (1 .. $max_n_cells) ### we chose SAMPLE_MATRIX random intensities from the matrix
		{
		my $i = int(rand($#$matrix_ref - 1)) + 1;
		my $j = int(rand($#{$$matrix_ref[$i]} - 1)) + 1;

		if (!exists($duplicated{$i}{$j}))
			{
			push(@sample_list, sprintf("%.2f", $$matrix_ref[$i][$j]));

			$duplicated{$i}{$j} = TRUE;

			$all_positive_flag = FALSE if ($$matrix_ref[$i][$j] < 0);
			$all_negative_flag = FALSE if ($$matrix_ref[$i][$j] > 0);
			}
		}
	

	if ($$user_arguments_ref{'VERBOSE_FLAG'} eq "TRUE")
		{
		print "[MATRIX] [WARNING] All sampled values in the matrix file are positive. Maybe argument 'center' should be set to TRUE?\n" if ($all_positive_flag == TRUE);
		print "[MATRIX] [WARNING] All sampled values in the matrix file are negative. Maybe argument 'center' should be set to TRUE?\n" if ($all_negative_flag == TRUE);
		}

	my @intensity_range_list;

	foreach my $value (@sample_list)
		{
		error($data_ref, "[ERROR] Matrix file has bad format") if (!defined($value));
		}

	error($data_ref, "[ERROR] Matrix file has no values <0 while argument 'center' is set to FALSE.") if (scalar(@sample_list) == 0);

	my $stat = Statistics::Descriptive::Full->new();
	$stat->add_data(@sample_list);

	my $min;
	my $max;

	if ($$user_arguments_ref{'INTENSITY'} =~ /IQR/)
		{
		my $q1 = $stat->percentile(25);
		my $q3 = $stat->percentile(75);
		my $iqr = ($q3 - $q1);
		$min = $q1 - (1.5*$iqr); 
		$max = $q3 + (1.5*$iqr); 
		}
	elsif ($$user_arguments_ref{'INTENSITY'} =~ /IDR/)
		{
		my $q1 = $stat->percentile(10);
		my $q3 = $stat->percentile(90);
		my $iqr = ($q3 - $q1);
		$min = $q1 - (1.5*$iqr); 
		$max = $q3 + (1.5*$iqr); 

		}
	elsif (looks_like_number($$user_arguments_ref{'INTENSITY'}))
		{
		$min = -1 * $$user_arguments_ref{'INTENSITY'};
		$max = $$user_arguments_ref{'INTENSITY'};
		}
	else
		{error($data_ref, "Error in calculating heatmap intensities.\n");}

	$max = 0 if ($all_negative_flag == TRUE);
	$min = 0 if ($all_positive_flag == TRUE);

	if ( ($all_negative_flag == FALSE) and ($all_positive_flag == FALSE) ) ### if there are both positive and negative values, centers the middle value to 0
			{
			$intensity_range_list[10] = 0;

			for my $i (1..10)
					{
					$intensity_range_list[10 + $i] = sprintf("%.3f", $i *  ($max / 10));
					}

			for my $i (1..10)
					{
					$intensity_range_list[10 - $i] = sprintf("%.3f", $i *  ($min / 10));
					}
			}
	elsif ($all_positive_flag == TRUE)
			{
			$intensity_range_list[10] = 0;

			for my $i (1..10)
					{
					$intensity_range_list[10 + $i] = sprintf("%.3f", $i *  ($max / 10));
					}

			for my $i (1..10)
					{
					$intensity_range_list[10 - $i] = 0;
					}
			}
	elsif ($all_negative_flag == TRUE)
			{
			$intensity_range_list[10] = 0;

			for my $i (1..10)
					{
					$intensity_range_list[10 + $i] = 0;
					}

			for my $i (1..10)
					{
					$intensity_range_list[10 - $i] = sprintf("%.3f", $i *  ($min / 10));
					}
			}
		

	return(@intensity_range_list);
	}


sub autocolors
	{
	my ($string, $list_ref) = @_;

	my @color_list = split(/\s+/, $string);

	my @random_colors_list = shuffle(@color_list);

	my %data;

	foreach my $value (@$list_ref)
		{
		$data{$value} = shift(@random_colors_list);

		if (scalar(@random_colors_list) == 0)
			{@random_colors_list = shuffle(@color_list);}
		}

	return(%data);
	}


sub display_version
	{
	print "
GETS (Gene Expression To Spreadsheet)
Version 1.0.0 (updated in September, 2019), Author: Daniel Aguilar\n";

	exit;
    }

sub display_help
	{
	print "
GETS (Gene Expression To Spreadsheet)
Version 1.0.0 (updated in September, 2019), Author: Daniel Aguilar

USAGE: $0 --matrix [--geneinfo] [--sampleinfo] [--output] [--colors] [--geneinfo-ordered] [--sampleinfo-ordered] [--sampleinfo-format] [--overwrite] [--verbose] [--center] [--intensity] [--heatmap-colors] [--palette] [--help] [--version]

OPTIONS (defaults in parentheses):
\tData files:
\t\t--matrix\tFile with expression matrix [required]
\t\t--geneinfo\tFile with gene information
\t\t--sampleinfo\tFile with sample information
\tColours and formats:
\t\t--colors\tFile with colors for gene and sample information
\t\t--intensity\tMethod for calculating the intensity in heatmap color gradient [IQR|IDR|<positive number larger than 0>] (IDR)
\t\t--heatmap-colors\tHeatmap color palette [GBR|BWR] (GBR)
\t\t--palette\tGene/sample information color palette [bright|muted] (bright)
\tMatrix pre-processing:
\t\t--center\tShould expression values be centered around the median? [TRUE|FALSE] (FALSE)
\tGene information pre-processing:
\t\t--geneinfo-ordered\tDoes row order in the gene information file match row order in the matrix file? [TRUE|FALSE] (TRUE)
\tSample information pre-processing:
\t\t--sampleinfo-ordered\tDoes column order in the sample information file match column order in the matrix file? [TRUE|FALSE] (TRUE)
\t\t--sampleinfo-format\tIs sample information provided as columns or as rows [columns|rows] (columns)
\tMiscellaneous:
\t\t--output\tPrefix for output files
\t\t--overwrite\tOverwrite output file? [TRUE|FALSE] (FALSE)
\t\t--verbose\tVerbose output? [TRUE|FALSE] (TRUE)

Check the documentation file for more information.
\n";

	exit;
	}


